#include "stdio.h"
#include "stdlib.h"
#include "string.h"
#include "conio.h"
#include "dos.h"

// Let's make sure we gracefully handle stuff(like) { this }!

#define QSIZE 50000
char line[150], queue[QSIZE], numstr[40], fspec[80];
unsigned qfront = 0, qrear = 0, qwrote_last = 0, after_extern = 65535;
int stack[100], sp = 0, num_files = 0,
    number_flag = 0, extern_flag = 0, no_statics = 0, statics_only = 0;

enum TextStatus
    {
    IN_QUESTION, IN_PARENS, AFTER_PARENS, AFTER_PARENS_MUTED, IN_BRACES,
    IN_COMMENT
    };

#define QUEUE_EMPTY()   ((qfront == qrear) && !qwrote_last)
#define QUEUE_FULL()    ((qfront == qrear) && qwrote_last)

char last_c = '\n';

void next_queue(unsigned *p)
    {
    (*p)++;
    if (*p == QSIZE) *p = 0;
    }

void kill_queue(void)
    {
    qfront = 0; qrear = 0; qwrote_last = 0;
    }

void flush_queue(void)
    {
    char c;

    if (QUEUE_EMPTY()) return;
    c = queue[qfront++];
    printf("%c", c);
    while (qfront != qrear)
        {
        c = queue[qfront];
        printf("%c", c);
        next_queue(&qfront);
        }
    qwrote_last = 0;
    after_extern = 65535;
    }

void enqueue(char c)
    {
    if (QUEUE_FULL())
        {
        kill_queue();
        printf("\n Problem! Queue full! %c\n", 7); 
        }
    qwrote_last = 1;
    queue[qrear] = c;
    next_queue(&qrear);
    last_c = c;
    }

void enqueue_string(char *s)
    {
    while (*s) enqueue(*s++);
    }

void setup_extern(void)
    {
    enqueue_string("extern ");
    after_extern = qrear;
    last_c = '\n';
    }

static int white_space(char c)
    {
    if (c == ' ') return 1;
    if (c == '\t') return 1;
    if (c == '\r') return 1;
    if (c == '\n') return 1;
    return 0;
    }

static unsigned skip_queue_white_space(unsigned i)
    {
    while (white_space(queue[i]))
        {
        next_queue(&i);
        if (i == qrear) return i;
        }
    return i;
    }

static unsigned queue_subcompare(unsigned i, char *s)
    {
    // compare string with queue contents starting at i
    while (*s != 0)
        {
        if (queue[i] != *s++) return 0;
        next_queue(&i);
        if (i == qrear) return 0;
        }
    return 1;
    }

static unsigned queue_compare(unsigned n, char *s)
    {
    unsigned i;

    if (QUEUE_EMPTY()) return 0;
    for (i = qfront; n > 0; next_queue(&i), n--)
        if (i == qrear) return 0;
    return queue_subcompare(skip_queue_white_space(i), s);
    }

static int statics_ok(void)
    {
    int is_static;

    if (!no_statics && !statics_only) return 1;
    // now I need to find out if it's actually a static
    is_static = queue_compare(0, "static");
    if (is_static && statics_only) return 1;
    if (!is_static && no_statics) return 1;
    return 0;
    }

void handle_one_file(char *fname)
    {
    unsigned i;
    char c;
    enum TextStatus status;
    long line_num;
    FILE *f;
    
    f = fopen(fname, "r");
    if (f == NULL) return;

    status = IN_QUESTION;
    line_num = 0;
    if (extern_flag) setup_extern();
    while (1)
        {
        if (fgets(line, 150, f) == 0)
            {
            kill_queue();
            fclose(f);
            return;
            }
        // get rid of the trailing carriage return
        line[strlen(line) - 1] = '\0';

        line_num++;

        if (strlen(line) < 2 || line[0] == '#') goto SkipLine;
        for (i = 0; i < strlen(line); i++)
            {
            c = line[i];
            switch (status)
                {
                case IN_QUESTION:
                    if (c == ';')
                        {
                        kill_queue();
                        if (extern_flag) setup_extern();
                        last_c = '\n';
                        break;
                        }
                    if (c == '/' && line[i+1] == '*')
                        {
                        stack[sp++] = status;
                        status = IN_COMMENT;
                        i++;
                        break;
                        }
                    if (c == '/' && line[i+1] == '/')
                        goto SkipLine;
                    if (c == '(')
                        {
                        enqueue(c);
                        stack[sp++] = status;
                        status = IN_PARENS;
                        break;
                        }
                    if (c == '{') /* Braces must match!! } */
                        {
                        stack[sp++] = status;
                        status = IN_BRACES;
                        break;
                        }
                    enqueue(c);
                    break;
                case IN_PARENS:
                    if (c == ')')
                        {
                        status = stack[--sp];
                        if (status == IN_QUESTION || status == AFTER_PARENS)
                            {
                            enqueue(')');
                            if (status == IN_QUESTION)
                                {
                                enqueue(';');
                                if (number_flag)
                                    {
                                    sprintf(numstr, " // %s %ld", fname, line_num);
                                    enqueue_string(numstr);
                                    }
                                }
                            enqueue('\n');
                            status = AFTER_PARENS;
                            }
                        break;
                        }
                    enqueue(c);
                    break;
                case AFTER_PARENS:
                    if (c == ';')
                        {
                        kill_queue();
                        status = IN_QUESTION;
                        break;
                        }
                    if (c == '/' && line[i+1] == '*')
                        {
                        stack[sp++] = status;
                        status = IN_COMMENT;
                        i++;
                        break;
                        }
                    if (c == '(')
                        {
                        sprintf(numstr,
                            "/* Probable syntax error, line %ld */ ",
                            line_num);
                        enqueue_string(numstr);
                        enqueue('(');
                        flush_queue();
                        stack[sp++] = status;
                        status = IN_PARENS;
                        break;
                        }
                    if (c == '{') /* Braces must match!! } */
                        {
                        if (extern_flag && after_extern != 65535 &&
                                queue_compare(strlen("extern "), "static"))
                            qfront = after_extern;
                        if (statics_ok()) flush_queue();
                        else kill_queue();
                        stack[sp++] = IN_QUESTION;
                        status = IN_BRACES;
                        if (extern_flag) setup_extern();
                        break;
                        }
                    break;
                case IN_BRACES:
                    if (c == '{')
                        {
                        stack[sp++] = status;
                        status = IN_BRACES;
                        break;
                        }
                    if (c == '}') status = stack[--sp];
                    break;
                case IN_COMMENT:
                    if (c == '*' && line[i+1] == '/')
                        {
                        status = stack[--sp];
                        i++;
                        }
                    break;
                }
            }
        if (last_c != '\n')
            switch (status)
                {
                case IN_QUESTION:
                    enqueue(' ');
                    last_c = '\n';
                    break;
                case AFTER_PARENS:
                case IN_PARENS:
                    enqueue('\n');
                    break;
                default:
                    break;
                }
SkipLine: ;
        }
    }

void print_help(void)
    {
    printf(
        "Usage: protos [switches] file-spec [file-spec2...]\n"
        "Input file specs may include '?' or '*' wildcards.\n"
        "Switches are optional, and may be merged, e.g. '-nge'\n"
        "Currently implemented flags are:\n"
        "-n => give line numbers for function definitions\n"
        "-e => add \"extern\" to prototypes\n"
        "-s => create prototypes only for static functions\n"
        "-g => create prototypes only for non-static (global) functions\n"
        "-f => get filespecs from a list (e.g. -ffilelist.txt)\n"
        );
    exit(0);
    }

void handle_file_spec(char *s)
    {
    unsigned j;
    struct find_t f;

    j = _dos_findfirst(s, _A_NORMAL, &f);
    while (j == 0)
        {
        handle_one_file(f.name);
        j = _dos_findnext(&f);
        num_files++;
        }
    }

void main(int argc, char *argv[])
    {
    unsigned i, j;
    FILE *f;
    
    if (argc == 1) print_help();
    for (i = 1; i < (unsigned) argc; i++)
        if (argv[i][0] == '-' || argv[i][0] == '/')
            for (j = 1; j < strlen(argv[i]); j++)
                switch (toupper(argv[i][j]))
                    {
                    case 'N':
                        number_flag = 1;
                        break;
                    case 'E':
                        extern_flag = 1;
                        break;
                    case 'S':
                        statics_only = 1;
                        no_statics = 0;
                        break;
                    case 'G':
                        statics_only = 0;
                        no_statics = 1;
                        break;
                    case 'F':
                        j = strlen(argv[i]);
                        break;
                    default:
                        break;
                    }

    for (i = 1; i < (unsigned) argc; i++)
        {
        if (argv[i][0] == '-' || argv[i][0] == '/')
            {
            for (j = 1; j < strlen(argv[i]); j++)
                if (toupper(argv[i][j]) == 'F')
                    {
                    f = fopen(&argv[i][j + 1], "r");
                    if (f == NULL) break;
                    while (!feof(f))
                        if (fgets(fspec, 80, f) != 0)
                            {
                            // get rid of the trailing carriage return
                            fspec[strlen(fspec) - 1] = '\0';
                            handle_file_spec(fspec);
                            }
                    fclose(f);
                    }
            }
        else handle_file_spec(argv[i]);
        }

    if (num_files == 0) { printf("No files found!\n"); exit(1); }
    exit(0);
    }
